import warnings

import cv2
import numpy as np
from scipy.ndimage import convolve
from scipy.ndimage import distance_transform_edt as bwdist
from skimage import measure, morphology

from .utils import EPS, TYPE, get_adaptive_threshold, validate_and_normalize_input


class Fmeasure(object):
    def __init__(self, beta: float = 0.3):
        """F-measure for SOD.

        ```
        @inproceedings{Fmeasure,
            title={Frequency-tuned salient region detection},
            author={Achanta, Radhakrishna and Hemami, Sheila and Estrada, Francisco and S{\"u}sstrunk, Sabine},
            booktitle=CVPR,
            number={CONF},
            pages={1597--1604},
            year={2009}
        }
        ```

        Args:
            beta (float): the weight of the precision
        """
        warnings.warn("This class will be removed in the future, please use FmeasureV2 instead!")

        self.beta = beta
        self.precisions = []
        self.recalls = []
        self.adaptive_fms = []
        self.changeable_fms = []

    def step(self, pred: np.ndarray, gt: np.ndarray, normalize: bool = True):
        """Statistics the metric for the pair of pred and gt.

        Args:
            pred (np.ndarray): Prediction, gray scale image.
            gt (np.ndarray): Ground truth, gray scale image.
            normalize (bool, optional): Whether to normalize the input data. Defaults to True.
        """
        pred, gt = validate_and_normalize_input(pred, gt, normalize)

        adaptive_fm = self.cal_adaptive_fm(pred=pred, gt=gt)
        self.adaptive_fms.append(adaptive_fm)

        precisions, recalls, changeable_fms = self.cal_pr(pred=pred, gt=gt)
        self.precisions.append(precisions)
        self.recalls.append(recalls)
        self.changeable_fms.append(changeable_fms)

    def cal_adaptive_fm(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """Calculate the adaptive F-measure.

        Returns:
            float: adaptive_fm
        """
        # ``np.count_nonzero`` is faster and better
        adaptive_threshold = get_adaptive_threshold(pred, max_value=1)
        binary_predcition = pred >= adaptive_threshold
        area_intersection = binary_predcition[gt].sum()
        if area_intersection == 0:
            adaptive_fm = 0
        else:
            pre = area_intersection / np.count_nonzero(binary_predcition)
            rec = area_intersection / np.count_nonzero(gt)
            adaptive_fm = (1 + self.beta) * pre * rec / (self.beta * pre + rec)
        return adaptive_fm

    def cal_pr(self, pred: np.ndarray, gt: np.ndarray) -> tuple:
        """Calculate the corresponding precision and recall when the threshold changes from 0 to 255.

        These precisions and recalls can be used to obtain the mean F-measure, maximum F-measure,
        precision-recall curve and F-measure-threshold curve.

        For convenience, `changeable_fms` is provided here, which can be used directly to obtain
        the mean F-measure, maximum F-measure and F-measure-threshold curve.

        Returns:
            tuple: (precisions, recalls, changeable_fms)
        """
        # 1. 获取预测结果在真值前背景区域中的直方图
        pred = (pred * 255).astype(np.uint8)
        bins = np.linspace(0, 256, 257)
        fg_hist, _ = np.histogram(pred[gt], bins=bins)  # 最后一个bin为[255, 256]
        bg_hist, _ = np.histogram(pred[~gt], bins=bins)
        # 2. 使用累积直方图（Cumulative Histogram）获得对应真值前背景中大于不同阈值的像素数量
        # 这里使用累加（cumsum）就是为了一次性得出 >=不同阈值 的像素数量, 这里仅计算了前景区域
        fg_w_thrs = np.cumsum(np.flip(fg_hist), axis=0)
        bg_w_thrs = np.cumsum(np.flip(bg_hist), axis=0)
        # 3. 使用不同阈值的结果计算对应的precision和recall
        # p和r的计算的真值是pred==1&gt==1，二者仅有分母不同，分母前者是pred==1，后者是gt==1
        # 为了同时计算不同阈值的结果，这里使用hsitogram&flip&cumsum 获得了不同各自的前景像素数量
        TPs = fg_w_thrs
        Ps = fg_w_thrs + bg_w_thrs
        # 为防止除0，这里针对除0的情况分析后直接对于0分母设为1，因为此时分子必为0
        Ps[Ps == 0] = 1
        T = max(np.count_nonzero(gt), 1)
        # TODO: T=0 或者 特定阈值下fg_w_thrs=0或者bg_w_thrs=0，这些都会包含在TPs[i]=0的情况中，
        #  但是这里使用TPs不便于处理列表
        precisions = TPs / Ps
        recalls = TPs / T

        numerator = (1 + self.beta) * precisions * recalls
        denominator = np.where(numerator == 0, 1, self.beta * precisions + recalls)
        changeable_fms = numerator / denominator
        return precisions, recalls, changeable_fms

    def get_results(self) -> dict:
        """Return the results about F-measure.

        Returns:
            dict(fm=dict(adp=adaptive_fm, curve=changeable_fm), pr=dict(p=precision, r=recall))
        """
        adaptive_fm = np.mean(np.array(self.adaptive_fms, TYPE))
        changeable_fm = np.mean(np.array(self.changeable_fms, dtype=TYPE), axis=0)
        precision = np.mean(np.array(self.precisions, dtype=TYPE), axis=0)  # N, 256
        recall = np.mean(np.array(self.recalls, dtype=TYPE), axis=0)  # N, 256
        return dict(fm=dict(adp=adaptive_fm, curve=changeable_fm), pr=dict(p=precision, r=recall))


class MAE(object):
    def __init__(self):
        """MAE(mean absolute error) for SOD.

        ```
        @inproceedings{MAE,
            title={Saliency filters: Contrast based filtering for salient region detection},
            author={Perazzi, Federico and Kr{\"a}henb{\"u}hl, Philipp and Pritch, Yael and Hornung, Alexander},
            booktitle=CVPR,
            pages={733--740},
            year={2012}
        }
        ```
        """
        self.maes = []

    def step(self, pred: np.ndarray, gt: np.ndarray, normalize: bool = True):
        """Statistics the metric for the pair of pred and gt.

        Args:
            pred (np.ndarray): Prediction, gray scale image.
            gt (np.ndarray): Ground truth, gray scale image.
            normalize (bool, optional): Whether to normalize the input data. Defaults to True.
        """
        pred, gt = validate_and_normalize_input(pred, gt, normalize)

        mae = self.cal_mae(pred, gt)
        self.maes.append(mae)

    def cal_mae(self, pred: np.ndarray, gt: np.ndarray) -> np.ndarray:
        """Calculate the mean absolute error.

        Returns:
            np.ndarray: mae
        """
        mae = np.mean(np.abs(pred - gt))
        return mae

    def get_results(self) -> dict:
        """Return the results about MAE.

        Returns:
            dict(mae=mae)
        """
        mae = np.mean(np.array(self.maes, TYPE))
        return dict(mae=mae)


class Smeasure(object):
    def __init__(self, alpha: float = 0.5):
        """S-measure(Structure-measure) of SOD.

        ```
        @inproceedings{Smeasure,
            title={Structure-measure: A new way to eval foreground maps},
            author={Fan, Deng-Ping and Cheng, Ming-Ming and Liu, Yun and Li, Tao and Borji, Ali},
            booktitle=ICCV,
            pages={4548--4557},
            year={2017}
        }
        ```

        Args:
            alpha: the weight for balancing the object score and the region score
        """
        self.sms = []
        self.alpha = alpha

    def step(self, pred: np.ndarray, gt: np.ndarray, normalize: bool = True):
        """Statistics the metric for the pair of pred and gt.

        Args:
            pred (np.ndarray): Prediction, gray scale image.
            gt (np.ndarray): Ground truth, gray scale image.
            normalize (bool, optional): Whether to normalize the input data. Defaults to True.
        """
        pred, gt = validate_and_normalize_input(pred, gt, normalize)

        sm = self.cal_sm(pred, gt)
        self.sms.append(sm)

    def cal_sm(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """Calculate the S-measure.

        Returns:
            s-measure
        """
        y = np.mean(gt)
        if y == 0:
            sm = 1 - np.mean(pred)
        elif y == 1:
            sm = np.mean(pred)
        else:
            object_score = self.object(pred, gt) * self.alpha
            region_score = self.region(pred, gt) * (1 - self.alpha)
            sm = max(0, object_score + region_score)
        return sm

    def s_object(self, x: np.ndarray) -> float:
        mean = np.mean(x)
        std = np.std(x, ddof=1)
        score = 2 * mean / (np.power(mean, 2) + 1 + std + EPS)
        return score

    def object(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """Calculate the object score."""
        gt_mean = np.mean(gt)
        fg_score = self.s_object(pred[gt]) * gt_mean
        bg_score = self.s_object((1 - pred)[~gt]) * (1 - gt_mean)
        object_score = fg_score + bg_score
        return object_score

    def region(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """Calculate the region score."""
        h, w = gt.shape
        area = h * w

        # Calculate the centroid coordinate of the foreground
        if np.count_nonzero(gt) == 0:
            cy, cx = np.round(h / 2), np.round(w / 2)
        else:
            # More details can be found at: https://www.yuque.com/lart/blog/gpbigm
            cy, cx = np.argwhere(gt).mean(axis=0).round()
        # To ensure consistency with the matlab code, one is added to the centroid coordinate,
        # so there is no need to use the redundant addition operation when dividing the region later,
        # because the sequence generated by ``1:X`` in matlab will contain ``X``.
        cy, cx = int(cy) + 1, int(cx) + 1

        # Use (x,y) to divide the ``pred`` and the ``gt`` into four submatrices, respectively.
        w_lt = cx * cy / area
        w_rt = cy * (w - cx) / area
        w_lb = (h - cy) * cx / area
        w_rb = 1 - w_lt - w_rt - w_lb
        score_lt = self.ssim(pred[0:cy, 0:cx], gt[0:cy, 0:cx]) * w_lt
        score_rt = self.ssim(pred[0:cy, cx:w], gt[0:cy, cx:w]) * w_rt
        score_lb = self.ssim(pred[cy:h, 0:cx], gt[cy:h, 0:cx]) * w_lb
        score_rb = self.ssim(pred[cy:h, cx:w], gt[cy:h, cx:w]) * w_rb
        return score_lt + score_rt + score_lb + score_rb

    def ssim(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """Calculate the ssim score."""
        h, w = pred.shape
        N = h * w

        x = np.mean(pred)
        y = np.mean(gt)

        sigma_x = np.sum((pred - x) ** 2) / (N - 1)
        sigma_y = np.sum((gt - y) ** 2) / (N - 1)
        sigma_xy = np.sum((pred - x) * (gt - y)) / (N - 1)

        alpha = 4 * x * y * sigma_xy
        beta = (x**2 + y**2) * (sigma_x + sigma_y)

        if alpha != 0:
            score = alpha / (beta + EPS)
        elif alpha == 0 and beta == 0:
            score = 1
        else:
            score = 0
        return score

    def get_results(self) -> dict:
        """Return the results about S-measure.

        Returns:
            dict(sm=sm)
        """
        sm = np.mean(np.array(self.sms, dtype=TYPE))
        return dict(sm=sm)


class Emeasure(object):
    def __init__(self):
        """E-measure(Enhanced-alignment Measure) for SOD.

        More details about the implementation can be found in https://www.yuque.com/lart/blog/lwgt38

        ```
        @inproceedings{Emeasure,
            title="Enhanced-alignment Measure for Binary Foreground Map Evaluation",
            author="Deng-Ping {Fan} and Cheng {Gong} and Yang {Cao} and Bo {Ren} and Ming-Ming {Cheng} and Ali {Borji}",
            booktitle=IJCAI,
            pages="698--704",
            year={2018}
        }
        ```
        """
        self.adaptive_ems = []
        self.changeable_ems = []

    def step(self, pred: np.ndarray, gt: np.ndarray, normalize: bool = True):
        """Statistics the metric for the pair of pred and gt.

        Args:
            pred (np.ndarray): Prediction, gray scale image.
            gt (np.ndarray): Ground truth, gray scale image.
            normalize (bool, optional): Whether to normalize the input data. Defaults to True.
        """
        pred, gt = validate_and_normalize_input(pred, gt, normalize)

        self.gt_fg_numel = np.count_nonzero(gt)
        self.gt_size = gt.shape[0] * gt.shape[1]

        changeable_ems = self.cal_changeable_em(pred, gt)
        self.changeable_ems.append(changeable_ems)
        adaptive_em = self.cal_adaptive_em(pred, gt)
        self.adaptive_ems.append(adaptive_em)

    def cal_adaptive_em(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """Calculate the adaptive E-measure.

        Returns:
            adaptive_em
        """
        adaptive_threshold = get_adaptive_threshold(pred, max_value=1)
        adaptive_em = self.cal_em_with_threshold(pred, gt, threshold=adaptive_threshold)
        return adaptive_em

    def cal_changeable_em(self, pred: np.ndarray, gt: np.ndarray) -> np.ndarray:
        """Calculate the changeable E-measure, which can be used to obtain the mean E-measure, the maximum E-measure and the E-measure-threshold curve.

        Returns:
            changeable_ems
        """
        changeable_ems = self.cal_em_with_cumsumhistogram(pred, gt)
        return changeable_ems

    def cal_em_with_threshold(self, pred: np.ndarray, gt: np.ndarray, threshold: float) -> float:
        """Calculate the E-measure corresponding to the specific threshold.

        Variable naming rules within the function:
        `[pred attribute(foreground fg, background bg)]_[gt attribute(foreground fg, background bg)]_[meaning]`

        If only `pred` or `gt` is considered, another corresponding attribute location is replaced with '`_`'.
        """
        binarized_pred = pred >= threshold
        fg_fg_numel = np.count_nonzero(binarized_pred & gt)
        fg_bg_numel = np.count_nonzero(binarized_pred & ~gt)

        fg___numel = fg_fg_numel + fg_bg_numel
        bg___numel = self.gt_size - fg___numel

        if self.gt_fg_numel == 0:
            enhanced_matrix_sum = bg___numel
        elif self.gt_fg_numel == self.gt_size:
            enhanced_matrix_sum = fg___numel
        else:
            parts_numel, combinations = self.generate_parts_numel_combinations(
                fg_fg_numel=fg_fg_numel,
                fg_bg_numel=fg_bg_numel,
                pred_fg_numel=fg___numel,
                pred_bg_numel=bg___numel,
            )

            results_parts = []
            for i, (part_numel, combination) in enumerate(zip(parts_numel, combinations)):
                align_matrix_value = (
                    2 * (combination[0] * combination[1]) / (combination[0] ** 2 + combination[1] ** 2 + EPS)
                )
                enhanced_matrix_value = (align_matrix_value + 1) ** 2 / 4
                results_parts.append(enhanced_matrix_value * part_numel)
            enhanced_matrix_sum = sum(results_parts)

        em = enhanced_matrix_sum / (self.gt_size - 1 + EPS)
        return em

    def cal_em_with_cumsumhistogram(self, pred: np.ndarray, gt: np.ndarray) -> np.ndarray:
        """Calculate the E-measure corresponding to the threshold that varies from 0 to 255..

        Variable naming rules within the function:
        `[pred attribute(foreground fg, background bg)]_[gt attribute(foreground fg, background bg)]_[meaning]`

        If only `pred` or `gt` is considered, another corresponding attribute location is replaced with '`_`'.
        """
        pred = (pred * 255).astype(np.uint8)
        bins = np.linspace(0, 256, 257)
        fg_fg_hist, _ = np.histogram(pred[gt], bins=bins)
        fg_bg_hist, _ = np.histogram(pred[~gt], bins=bins)
        fg_fg_numel_w_thrs = np.cumsum(np.flip(fg_fg_hist), axis=0)
        fg_bg_numel_w_thrs = np.cumsum(np.flip(fg_bg_hist), axis=0)

        fg___numel_w_thrs = fg_fg_numel_w_thrs + fg_bg_numel_w_thrs
        bg___numel_w_thrs = self.gt_size - fg___numel_w_thrs

        if self.gt_fg_numel == 0:
            enhanced_matrix_sum = bg___numel_w_thrs
        elif self.gt_fg_numel == self.gt_size:
            enhanced_matrix_sum = fg___numel_w_thrs
        else:
            parts_numel_w_thrs, combinations = self.generate_parts_numel_combinations(
                fg_fg_numel=fg_fg_numel_w_thrs,
                fg_bg_numel=fg_bg_numel_w_thrs,
                pred_fg_numel=fg___numel_w_thrs,
                pred_bg_numel=bg___numel_w_thrs,
            )

            results_parts = np.empty(shape=(4, 256), dtype=np.float64)
            for i, (part_numel, combination) in enumerate(zip(parts_numel_w_thrs, combinations)):
                align_matrix_value = (
                    2 * (combination[0] * combination[1]) / (combination[0] ** 2 + combination[1] ** 2 + EPS)
                )
                enhanced_matrix_value = (align_matrix_value + 1) ** 2 / 4
                results_parts[i] = enhanced_matrix_value * part_numel
            enhanced_matrix_sum = results_parts.sum(axis=0)

        em = enhanced_matrix_sum / (self.gt_size - 1 + EPS)
        return em

    def generate_parts_numel_combinations(self, fg_fg_numel, fg_bg_numel, pred_fg_numel, pred_bg_numel):
        bg_fg_numel = self.gt_fg_numel - fg_fg_numel
        bg_bg_numel = pred_bg_numel - bg_fg_numel

        parts_numel = [fg_fg_numel, fg_bg_numel, bg_fg_numel, bg_bg_numel]

        mean_pred_value = pred_fg_numel / self.gt_size
        mean_gt_value = self.gt_fg_numel / self.gt_size

        demeaned_pred_fg_value = 1 - mean_pred_value
        demeaned_pred_bg_value = 0 - mean_pred_value
        demeaned_gt_fg_value = 1 - mean_gt_value
        demeaned_gt_bg_value = 0 - mean_gt_value

        combinations = [
            (demeaned_pred_fg_value, demeaned_gt_fg_value),
            (demeaned_pred_fg_value, demeaned_gt_bg_value),
            (demeaned_pred_bg_value, demeaned_gt_fg_value),
            (demeaned_pred_bg_value, demeaned_gt_bg_value),
        ]
        return parts_numel, combinations

    def get_results(self) -> dict:
        """Return the results about E-measure.

        Returns:
            dict(em=dict(adp=adaptive_em, curve=changeable_em))
        """
        adaptive_em = np.mean(np.array(self.adaptive_ems, dtype=TYPE))
        changeable_em = np.mean(np.array(self.changeable_ems, dtype=TYPE), axis=0)
        return dict(em=dict(adp=adaptive_em, curve=changeable_em))


class WeightedFmeasure(object):
    def __init__(self, beta: float = 1):
        """Weighted F-measure for SOD.

        ```
        @inproceedings{wFmeasure,
            title={How to eval foreground maps?},
            author={Margolin, Ran and Zelnik-Manor, Lihi and Tal, Ayellet},
            booktitle=CVPR,
            pages={248--255},
            year={2014}
        }
        ```

        Args:
            beta (float): the weight of the precision
        """
        self.beta = beta
        self.weighted_fms = []

    def step(self, pred: np.ndarray, gt: np.ndarray, normalize: bool = True):
        """Statistics the metric for the pair of pred and gt.

        Args:
            pred (np.ndarray): Prediction, gray scale image.
            gt (np.ndarray): Ground truth, gray scale image.
            normalize (bool, optional): Whether to normalize the input data. Defaults to True.
        """
        pred, gt = validate_and_normalize_input(pred, gt, normalize)

        if np.all(~gt):
            wfm = 0
        else:
            wfm = self.cal_wfm(pred, gt)
        self.weighted_fms.append(wfm)

    def cal_wfm(self, pred: np.ndarray, gt: np.ndarray) -> float:
        """Calculate the weighted F-measure."""
        # [Dst,IDXT] = bwdist(dGT);
        Dst, Idxt = bwdist(gt == 0, return_indices=True)

        # %Pixel dependency
        # E = abs(FG-dGT);
        E = np.abs(pred - gt)
        # Et = E;
        # Et(~GT)=Et(IDXT(~GT)); %To deal correctly with the edges of the foreground region
        Et = np.copy(E)
        Et[gt == 0] = Et[Idxt[0][gt == 0], Idxt[1][gt == 0]]

        # K = fspecial('gaussian',7,5);
        # EA = imfilter(Et,K);
        K = self.matlab_style_gauss2D((7, 7), sigma=5)
        EA = convolve(Et, weights=K, mode="constant", cval=0)
        # MIN_E_EA = E;
        # MIN_E_EA(GT & EA<E) = EA(GT & EA<E);
        MIN_E_EA = np.where(gt & (EA < E), EA, E)

        # %Pixel importance
        # B = ones(size(GT));
        # B(~GT) = 2-1*exp(log(1-0.5)/5.*Dst(~GT));
        # Ew = MIN_E_EA.*B;
        B = np.where(gt == 0, 2 - np.exp(np.log(0.5) / 5 * Dst), np.ones_like(gt))
        Ew = MIN_E_EA * B

        # TPw = sum(dGT(:)) - sum(sum(Ew(GT)));
        # FPw = sum(sum(Ew(~GT)));
        TPw = np.sum(gt) - np.sum(Ew[gt == 1])
        FPw = np.sum(Ew[gt == 0])

        # R = 1- mean2(Ew(GT)); %Weighed Recall
        # P = TPw./(eps+TPw+FPw); %Weighted Precision
        # 注意这里使用mask索引矩阵的时候不可使用Ew[gt]，这实际上仅在索引Ew的0维度
        R = 1 - np.mean(Ew[gt == 1])
        P = TPw / (TPw + FPw + EPS)

        # % Q = (1+Beta^2)*(R*P)./(eps+R+(Beta.*P));
        Q = (1 + self.beta) * R * P / (R + self.beta * P + EPS)

        return Q

    def matlab_style_gauss2D(self, shape: tuple = (7, 7), sigma: int = 5) -> np.ndarray:
        """2D gaussian mask - should give the same result as MATLAB's:
        `fspecial('gaussian',[shape],[sigma])`
        """
        m, n = [(ss - 1) / 2 for ss in shape]
        y, x = np.ogrid[-m : m + 1, -n : n + 1]
        h = np.exp(-(x * x + y * y) / (2 * sigma * sigma))
        h[h < np.finfo(h.dtype).eps * h.max()] = 0
        sumh = h.sum()
        if sumh != 0:
            h /= sumh
        return h

    def get_results(self) -> dict:
        """Return the results about weighted F-measure.

        Returns:
            dict(wfm=weighted_fm)
        """
        weighted_fm = np.mean(np.array(self.weighted_fms, dtype=TYPE))
        return dict(wfm=weighted_fm)


class HumanCorrectionEffortMeasure(object):
    def __init__(self, relax: int = 5, epsilon: float = 2.0):
        """Human Correction Effort Measure for Dichotomous Image Segmentation.

        ```
        @inproceedings{HumanCorrectionEffortMeasure,
            title = {Highly Accurate Dichotomous Image Segmentation},
            author = {Xuebin Qin and Hang Dai and Xiaobin Hu and Deng-Ping Fan and Ling Shao and Luc Van Gool},
            booktitle = ECCV,
            year = {2022}
        }
        ```
        """

        self.hces = []
        self.relax = relax
        self.epsilon = epsilon
        self.morphology_kernel = morphology.disk(1)

    def step(self, pred: np.ndarray, gt: np.ndarray, normalize: bool = True):
        """Statistics the metric for the pair of pred and gt.

        Args:
            pred (np.ndarray): Prediction, gray scale image.
            gt (np.ndarray): Ground truth, gray scale image.
            normalize (bool, optional): Whether to normalize the input data. Defaults to True.
        """
        pred, gt = validate_and_normalize_input(pred, gt, normalize)

        hce = self.cal_hce(pred, gt)
        self.hces.append(hce)

    def cal_hce(self, pred: np.ndarray, gt: np.ndarray) -> float:
        gt_skeleton = morphology.skeletonize(gt).astype(bool)
        pred = pred > 0.5

        union = np.logical_or(gt, pred)
        TP = np.logical_and(gt, pred)
        FP = np.logical_xor(pred, TP)
        FN = np.logical_xor(gt, TP)

        # relax the union of gt and pred
        eroded_union = cv2.erode(union.astype(np.uint8), self.morphology_kernel, iterations=self.relax)

        # get the relaxed FP regions for computing the human efforts in correcting them ---
        FP_ = np.logical_and(FP, eroded_union)  # get the relaxed FP
        for i in range(0, self.relax):
            FP_ = cv2.dilate(FP_.astype(np.uint8), self.morphology_kernel)
            FP_ = np.logical_and(FP_.astype(bool), ~gt)
        FP_ = np.logical_and(FP, FP_)

        # get the relaxed FN regions for computing the human efforts in correcting them ---
        FN_ = np.logical_and(FN, eroded_union)  # preserve the structural components of FN
        # recover the FN, where pixels are not close to the TP borders
        for i in range(0, self.relax):
            FN_ = cv2.dilate(FN_.astype(np.uint8), self.morphology_kernel)
            FN_ = np.logical_and(FN_, ~pred)
        FN_ = np.logical_and(FN, FN_)
        # preserve the structural components of FN
        FN_ = np.logical_or(FN_, np.logical_xor(gt_skeleton, np.logical_and(TP, gt_skeleton)))

        # Find exact polygon control points and independent regions.
        # find contours from FP_ and control points and independent regions for human correction
        contours_FP, _ = cv2.findContours(FP_.astype(np.uint8), cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
        condition_FP = np.logical_or(TP, FN_)
        bdies_FP, indep_cnt_FP = self.filter_conditional_boundary(contours_FP, FP_, condition_FP)
        # find contours from FN_ and control points and independent regions for human correction
        contours_FN, _ = cv2.findContours(FN_.astype(np.uint8), cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)
        condition_FN = 1 - np.logical_or(np.logical_or(TP, FP_), FN_)
        bdies_FN, indep_cnt_FN = self.filter_conditional_boundary(contours_FN, FN_, condition_FN)

        poly_FP_point_cnt = self.count_polygon_control_points(bdies_FP, epsilon=self.epsilon)
        poly_FN_point_cnt = self.count_polygon_control_points(bdies_FN, epsilon=self.epsilon)
        return poly_FP_point_cnt + indep_cnt_FP + poly_FN_point_cnt + indep_cnt_FN

    def filter_conditional_boundary(self, contours: list, mask: np.ndarray, condition: np.ndarray):
        """
        Filter boundary segments based on a given condition mask and compute
        the number of independent connected regions that require human correction.

        Args:
            contours (List[np.ndarray]): List of boundary contours (OpenCV format).
            mask (np.ndarray): Binary mask representing the region of interest.
            condition (np.ndarray): Condition mask used to determine which
                boundary points need to be considered.

        Returns:
            Tuple[List[np.ndarray], int]:
                - boundaries (List[np.ndarray]): Filtered boundary segments that require correction.
                - independent_count (int): Number of independent connected regions
                that need correction (i.e., human editing effort).
        """
        condition = cv2.dilate(condition.astype(np.uint8), self.morphology_kernel)

        labels = measure.label(mask)  # find the connected regions
        independent_flags = np.ones(labels.max() + 1, dtype=int)  # the label of each connected regions
        independent_flags[0] = 0  # 0 indicate the background region

        boundaries = []
        visited_map = np.zeros(condition.shape[:2], dtype=int)
        for i in range(len(contours)):
            temp_boundaries = []
            temp_boundary = []
            for pt in contours[i]:
                row, col = pt[0, 1], pt[0, 0]

                if condition[row, col].sum() == 0 or visited_map[row, col] != 0:
                    if temp_boundary:  # if the previous point is not a boundary point, append the previous boundary
                        temp_boundaries.append(temp_boundary)
                        temp_boundary = []
                    continue

                temp_boundary.append([col, row])
                visited_map[row, col] = visited_map[row, col] + 1
                independent_flags[labels[row, col]] = 0  # mark region as requiring correction

            if temp_boundary:
                temp_boundaries.append(temp_boundary)

            # check if the first and the last boundaries are connected.
            # if yes, invert the first boundary and attach it after the last boundary
            if len(temp_boundaries) > 1:
                first_x, first_y = temp_boundaries[0][0]
                last_x, last_y = temp_boundaries[-1][-1]
                if (
                    (abs(first_x - last_x) == 1 and first_y == last_y)
                    or (first_x == last_x and abs(first_y - last_y) == 1)
                    or (abs(first_x - last_x) == 1 and abs(first_y - last_y) == 1)
                ):
                    temp_boundaries[-1].extend(temp_boundaries[0][::-1])
                    del temp_boundaries[0]

            for k in range(len(temp_boundaries)):
                temp_boundaries[k] = np.array(temp_boundaries[k])[:, np.newaxis, :]

            if temp_boundaries:
                boundaries.extend(temp_boundaries)
        return boundaries, independent_flags.sum()

    def count_polygon_control_points(self, boundaries: list, epsilon: float = 1.0) -> int:
        """
        Approximate each boundary using the Ramer-Douglas-Peucker (RDP) algorithm
        and count the total number of control points of all approximated polygons.

        Args:
            boundaries (List[np.ndarray]): List of boundary contours.
                Each contour is an Nx1x2 numpy array (OpenCV contour format).
            epsilon (float): RDP approximation tolerance.
                Larger values result in fewer control points.

        Returns:
            int: The total number of control points across all approximated polygons.

        Reference:
            https://en.wikipedia.org/wiki/Ramer-Douglas-Peucker_algorithm
        """
        num_points = 0
        for boundary in boundaries:
            approx_poly = cv2.approxPolyDP(boundary, epsilon, False)  # approximate boundary
            num_points += len(approx_poly)  # count vertices (control points)
        return num_points

    def get_results(self) -> dict:
        hce = np.mean(np.array(self.hces, dtype=TYPE))
        return dict(hce=hce)



import numpy as np

# the different implementation of epsilon (extreme min value) between numpy and matlab
EPS = np.spacing(1)
TYPE = np.float64


def validate_and_normalize_input(pred: np.ndarray, gt: np.ndarray, normalize: bool = True):
    """Performs input validation and normalization."""
    # Validate input shapes
    if pred.shape != gt.shape:
        raise ValueError(
            f"Shape mismatch between prediction ({pred.shape}) and ground truth ({gt.shape})"
        )

    # Handle normalization
    if normalize:
        pred, gt = prepare_data(pred, gt)
    else:
        # Validate prediction data type and range
        if pred.dtype not in (np.float32, np.float64):
            raise TypeError(f"Prediction array must be float32 or float64, got {pred.dtype}")
        if not (0 <= pred.min() and pred.max() <= 1):
            raise ValueError("Prediction values must be in range [0, 1]")
        # Validate ground truth type
        if gt.dtype != bool:
            raise TypeError(f"Ground truth must be boolean, got {gt.dtype}")

    return pred, gt


def prepare_data(pred: np.ndarray, gt: np.ndarray) -> tuple:
    """A numpy-based function for preparing `pred` and `gt`.

    - for `pred`, it looks like `mapminmax(im2double(...))` of matlab;
    - `gt` will be binarized by 128.

    Args:
        pred (np.uint8): Prediction, gray scale image.
        gt (np.uint8): Ground truth, gray scale image.

    Returns:
        tuple: pred (np.float64), gt (bool)
    """
    gt = gt > 128
    # im2double, mapminmax
    pred = pred / 255
    if pred.max() != pred.min():
        pred = (pred - pred.min()) / (pred.max() - pred.min())
    return pred, gt


def get_adaptive_threshold(matrix: np.ndarray, max_value: float = 1) -> float:
    """Return an adaptive threshold, which is equal to twice the mean of `matrix`.

    Args:
        matrix (np.ndarray): a data array
        max_value (float, optional): the upper limit of the threshold. Defaults to 1.

    Returns:
        float: `min(2 * matrix.mean(), max_value)`
    """
    return min(2 * matrix.mean(), max_value)



import numpy as np
from scipy import ndimage

from .utils import TYPE, get_adaptive_threshold, validate_and_normalize_input


class MSIoU:
    """
    Multi-Scale Intersection over Union (MSIoU) metric.

    ::

        @inproceedings{MSIoU,
            title = {Multiscale IOU: A Metric for Evaluation of Salient Object Detection with Fine Structures},
            author = {Ahmadzadeh, Azim and Kempton, Dustin J. and Chen, Yang and Angryk, Rafal A.},
            booktitle = ICIP,
            year = {2021},
        }
    """

    def __init__(
        self,
        with_dynamic: bool,
        with_adaptive: bool,
        *,
        with_binary: bool = False,
        num_levels=10,
    ):
        """
        Args:
            with_dynamic (bool, optional): Record dynamic results for max/avg/curve versions.
            with_adaptive (bool, optional): Record adaptive results for adp version.
            with_binary (bool, optional): Record binary results for binary version.
        """
        self.dynamic_results = [] if with_dynamic else None
        self.adaptive_results = [] if with_adaptive else None
        self.binary_results = [] if with_binary else None

        # The values of this collection determines the resolutions based on which MIoU is computed.
        # It is set as the original implementation
        self.cell_sizes = np.power(2, np.linspace(0, 9, num=num_levels, dtype=int))

    def get_edge(self, mask: np.ndarray):
        """Edge detection based on the `scipy.ndimage.sobel` function.

        :param mask: a binary mask of an object whose edges are of interest.
        :return: a binary mask of 1's as edges and 0's as background.
        """
        sx = ndimage.sobel(mask, axis=0, mode="constant")
        sy = ndimage.sobel(mask, axis=1, mode="constant")
        sob = np.hypot(sx, sy)
        # sob[sob > 0] = 1
        return (sob > 0).astype(sob.dtype)

    def shrink_by_grid(self, image: np.ndarray, cell_size: int) -> np.ndarray:
        """Shrink the image by summing values within grid cells.

        Performs box-counting after applying zero padding if the image dimensions
        are not perfectly divisible by the cell size.

        :param image: The input binary image (edges).
        :param cell_size: The size of the grid cells.
        :return: A shrunk binary image where each pixel represents a grid cell.
        """
        if cell_size <= 0:
            raise ValueError("Cell size must be a positive integer")

        if cell_size > 1:
            # Calculate padding sizes to make dimensions divisible by cell_size
            h, w = image.shape[:2]
            pad_h = (cell_size - h % cell_size) % cell_size
            pad_w = (cell_size - w % cell_size) % cell_size

            # Apply padding if necessary
            if pad_h > 0 or pad_w > 0:
                # Padding is added to the top and left edges.
                image = np.pad(image, ((pad_h, 0), (pad_w, 0)), mode="constant", constant_values=0)

            # Reshape and sum within each cell
            h, w = image.shape[:2]
            image = image.reshape(h // cell_size, cell_size, w // cell_size, cell_size)
            image = image.sum(axis=(1, 3))
        # image[image > 0] = 1
        return (image > 0).astype(image.dtype)

    def multi_scale_iou(self, pred_edge: np.ndarray, gt_edge: np.ndarray) -> list:
        """Calculate Multi-Scale IoU.

        Args:
            pred_edge (np.ndarray): edge map of pred
            gt_edge (np.ndarray): edge map of gt

        Returns:
            list: ratios
        """

        # Calculate IoU ratios at different scales
        ratios = []
        for cell_size in self.cell_sizes:
            # Shrink both prediction and ground truth edges
            shrunk_pred_edge = self.shrink_by_grid(pred_edge, cell_size=cell_size)
            shrunk_gt_edge = self.shrink_by_grid(gt_edge, cell_size=cell_size)

            # Calculate IoU with smoothing to prevent division by zero
            numerator = np.logical_and(shrunk_pred_edge, shrunk_gt_edge).sum() + 1
            # Only consider ground truth for denominator
            denominator = shrunk_gt_edge.sum() + 1
            ratios.append(numerator / denominator)
        return ratios

    def binarizing(self, pred_bin: np.ndarray, gt_edge: np.ndarray) -> list:
        """Calculate Multi-Scale IoU based on dynamically thresholding.

        Args:
            pred_bin (np.ndarray): binarized pred
            gt_edge (np.ndarray): gt binarized by 128

        Returns:
            np.ndarray: areas under the curve
        """
        pred_edge = self.get_edge(pred_bin)
        ratios = self.multi_scale_iou(pred_edge, gt_edge)  # 10

        # Calculate area under the curve using trapezoidal rule
        return np.trapz(y=ratios, dx=1 / (len(self.cell_sizes) - 1))

    def step(self, pred: np.ndarray, gt: np.ndarray, normalize: bool = True):
        """Calculate the Multi-Scale IoU for a single prediction-ground truth pair.

        This method first extracts edges from both prediction and ground truth,
        then computes IoU ratios at multiple scales defined by self.cell_sizes.
        Finally, it calculates the area under the curve of these ratios.

        Args:
            pred (np.ndarray): Prediction, gray scale image.
            gt (np.ndarray): Ground truth, gray scale image.
            normalize (bool, optional): Whether to normalize the input data. Defaults to True.

        Returns:
            The MSIoU score for the given pair (float between 0 and 1).
        """
        pred, gt = validate_and_normalize_input(pred, gt, normalize)

        # Calculate MSIoU for this pair and store the result
        gt_edge = self.get_edge(gt)

        if self.dynamic_results is not None:
            results = []
            _pred = (pred * 255).astype(np.uint8)
            for threshold in np.linspace(0, 256, 257):
                results.append(self.binarizing(_pred >= threshold, gt_edge))
            # threshold_masks = pred[..., None] >= np.arange(0, 257)[None, None, :]
            self.dynamic_results.append(results)

        if self.adaptive_results is not None:
            adaptive_threshold = get_adaptive_threshold(pred, max_value=1)
            results = self.binarizing(pred >= adaptive_threshold, gt_edge)
            self.adaptive_results.append(results)

        if self.binary_results is not None:
            self.binary_results.append(self.binarizing(pred > 0.5, gt_edge))

    def get_results(self) -> dict:
        """Return the results about MSIoU.

        Calculates the mean of all stored MSIoU values from previous calls to step().

        :return: Dictionary with key 'msiou' and the mean MSIoU value.
        :raises: ValueError if no samples have been processed.
        """
        results = {}
        if self.dynamic_results is not None:
            results["dynamic"] = np.mean(np.array(self.dynamic_results, dtype=TYPE), axis=0)
        if self.adaptive_results is not None:
            results["adaptive"] = np.mean(np.array(self.adaptive_results, dtype=TYPE))
        if self.binary_results is not None:
            results["binary"] = np.mean(np.array(self.binary_results, dtype=TYPE))
        return results



# -*- coding: utf-8 -*-
import abc

import numpy as np

from .utils import TYPE, get_adaptive_threshold, validate_and_normalize_input


class _BaseHandler:
    """Base class for all metric handlers.

    Provides common functionality for calculating various segmentation metrics.
    """

    def __init__(
        self,
        with_dynamic: bool,
        with_adaptive: bool,
        *,
        with_binary: bool = False,
        sample_based: bool = True,
    ):
        """
        Args:
            with_dynamic (bool, optional): Record dynamic results for max/avg/curve versions.
            with_adaptive (bool, optional): Record adaptive results for adp version.
            with_binary (bool, optional): Record binary results for binary version.
            sample_based (bool, optional): Whether to average the metric of each sample or calculate
                the metric of the dataset. Defaults to True.
        """
        self.dynamic_results = [] if with_dynamic else None
        self.adaptive_results = [] if with_adaptive else None
        self.sample_based = sample_based
        if with_binary:
            if self.sample_based:
                self.binary_results = []
            else:
                self.binary_results = {"tp": 0, "fp": 0, "tn": 0, "fn": 0}
        else:
            self.binary_results = None

    @abc.abstractmethod
    def __call__(self, tp, fp, tn, fn):
        """Calculate the metric value.

        Args:
            tp: True positive count(s)
            fp: False positive count(s)
            tn: True negative count(s)
            fn: False negative count(s)

        Returns:
            Calculated metric value(s)
        """
        pass

    @staticmethod
    def divide(numerator, denominator):
        """Safe division with numpy arrays handling zero denominators.

        Args:
            numerator: Numerator value(s)
            denominator: Denominator value(s)

        Returns:
            Result of division with zero handling
        """
        denominator = np.array(denominator, dtype=TYPE)
        np.divide(numerator, denominator, out=denominator, where=denominator != 0)
        return denominator


class IOUHandler(_BaseHandler):
    """Intersection over Union

    iou = tp / (tp + fp + fn)
    """

    def __call__(self, tp, fp, tn, fn):
        # ious = np.where(Ps + FNs == 0, 0, TPs / (Ps + FNs))
        return self.divide(tp, tp + fp + fn)


class SpecificityHandler(_BaseHandler):
    """Specificity

    True negative rate (TNR)/specificity (SPC)/selectivity

    specificity = tn / (tn + fp)
    """

    def __call__(self, tp, fp, tn, fn):
        # specificities = np.where(TNs + FPs == 0, 0, TNs / (TNs + FPs))
        return self.divide(tn, tn + fp)


TNRHandler = SpecificityHandler


class DICEHandler(_BaseHandler):
    """DICE

    dice = 2 * tp / (tp + fn + tp + fp)
    """

    def __call__(self, tp, fp, tn, fn):
        # dices = np.where(TPs + FPs == 0, 0, 2 * TPs / (T + Ps))
        return self.divide(2 * tp, tp + fn + tp + fp)


class OverallAccuracyHandler(_BaseHandler):
    """OverallAccuracy

    oa = overall_accuracy = (tp + tn) / (tp + fp + tn + fn)
    """

    def __call__(self, tp, fp, tn, fn):
        # dices = np.where(TPs + FPs == 0, 0, 2 * TPs / (T + Ps))
        return self.divide(tp + tn, tp + fp + tn + fn)


class KappaHandler(_BaseHandler):
    """KappaAccuracy

    kappa = kappa = (oa - p_) / (1 - p_)
    p_ = [(tp + fp)(tp + fn) + (tn + fn)(tn + tp)] / (tp + fp + tn + fn)^2
    """

    def __init__(
        self,
        with_dynamic: bool,
        with_adaptive: bool,
        *,
        with_binary: bool = False,
        sample_based: bool = True,
    ):
        """
        Args:
            with_dynamic (bool, optional): Record dynamic results for max/avg/curve versions.
            with_adaptive (bool, optional): Record adaptive results for adp version.
            with_binary (bool, optional): Record binary results for binary version.
            sample_based (bool, optional): Whether to average the metric of each sample or calculate
                the metric of the dataset. Defaults to True.
        """
        super().__init__(
            with_dynamic=with_dynamic,
            with_adaptive=with_adaptive,
            with_binary=with_binary,
            sample_based=sample_based,
        )

        self.oa = OverallAccuracyHandler(False, False)

    def __call__(self, tp, fp, tn, fn):
        oa = self.oa(tp, fp, tn, fn)
        hpy_p = self.divide(
            (tp + fp) * (tp + fn) + (tn + fn) * (tn + tp),
            (tp + fp + tn + fn) ** 2,
        )
        return self.divide(oa - hpy_p, 1 - hpy_p)


class PrecisionHandler(_BaseHandler):
    """Precision

    precision = tp / (tp + fp)
    """

    def __call__(self, tp, fp, tn, fn):
        # precisions = np.where(Ps == 0, 0, TPs / Ps)
        return self.divide(tp, tp + fp)


class RecallHandler(_BaseHandler):
    """Recall

    True positive rate (TPR)/recall/sensitivity (SEN)/probability of detection/hit rate/power

    recall = tp / (tp + fn)
    """

    def __call__(self, tp, fp, tn, fn):
        # recalls = np.where(TPs == 0, 0, TPs / T)
        return self.divide(tp, tp + fn)


TPRHandler = RecallHandler
SensitivityHandler = RecallHandler


class FPRHandler(_BaseHandler):
    """False Positive Rate

    False positive rate (FPR)/probability of false alarm/fall-out

    fpr = fp / (tn + fp)
    """

    def __call__(self, tp, fp, tn, fn):
        return self.divide(fp, tn + fp)


class BERHandler(_BaseHandler):
    """Balance Error Rate

    ber = 1 - 0.5 * (tp / (tp + fn) + tn / (tn + fp))
    """

    def __call__(self, tp, fp, tn, fn):
        fg = np.asarray(tp + fn, dtype=TYPE)
        bg = np.asarray(tn + fp, dtype=TYPE)
        np.divide(tp, fg, out=fg, where=fg != 0)
        np.divide(tn, bg, out=bg, where=bg != 0)
        return 1 - 0.5 * (fg + bg)


class FmeasureHandler(_BaseHandler):
    """F-measure

    fmeasure = (beta + 1) * precision * recall / (beta * precision + recall)
    """

    def __init__(
        self,
        with_dynamic: bool,
        with_adaptive: bool,
        *,
        with_binary: bool = False,
        sample_based: bool = True,
        beta: float = 0.3,
    ):
        """
        Args:
            with_dynamic (bool, optional): Record dynamic results for max/avg/curve versions.
            with_adaptive (bool, optional): Record adaptive results for adp version.
            with_binary (bool, optional): Record binary results for binary version.
            sample_based (bool, optional): Whether to average the metric of each sample or calculate
                the metric of the dataset. Defaults to True.
            beta (bool, optional): β^2 in F-measure. Defaults to 0.3.
        """
        super().__init__(
            with_dynamic=with_dynamic,
            with_adaptive=with_adaptive,
            with_binary=with_binary,
            sample_based=sample_based,
        )

        self.beta = beta
        self.precision = PrecisionHandler(False, False)
        self.recall = RecallHandler(False, False)

    def __call__(self, tp, fp, tn, fn):
        """
        Note:
            Uses separate precision and recall calculations to maintain consistency with original implementation rather than combined formula.
        """
        p = self.precision(tp, fp, tn, fn)
        r = self.recall(tp, fp, tn, fn)
        return self.divide((self.beta + 1) * p * r, self.beta * p + r)


class FmeasureV2:
    def __init__(self, metric_handlers: dict = None):
        """Enhanced Fmeasure class with more relevant metrics, e.g. precision, recall, specificity, dice, iou, fmeasure and so on.

        Args:
            metric_handlers (dict, optional): Handlers of different metrics. Defaults to None.
        """
        self._metric_handlers = metric_handlers if metric_handlers else {}

    def add_handler(self, handler_name, metric_handler):
        self._metric_handlers[handler_name] = metric_handler

    @staticmethod
    def get_statistics(binary: np.ndarray, gt: np.ndarray, FG: int, BG: int) -> dict:
        """Calculate the TP, FP, TN and FN based a adaptive threshold.

        Args:
            binary (np.ndarray): binarized `pred` containing [0, 1]
            gt (np.ndarray): gt binarized by 128
            FG (int): the number of foreground pixels in gt
            BG (int): the number of background pixels in gt

        Returns:
            dict: TP, FP, TN, FN
        """
        TP = np.count_nonzero(binary[gt])
        FP = np.count_nonzero(binary[~gt])
        FN = FG - TP
        TN = BG - FP
        return {"tp": TP, "fp": FP, "tn": TN, "fn": FN}

    def adaptively_binarizing(self, pred: np.ndarray, gt: np.ndarray, FG: int, BG: int) -> dict:
        """Calculate the TP, FP, TN and FN based a adaptive threshold.

        Args:
            pred (np.ndarray): prediction normalized in [0, 1]
            gt (np.ndarray): gt binarized by 128
            FG (int): the number of foreground pixels in gt
            BG (int): the number of background pixels in gt

        Returns:
            dict: TP, FP, TN, FN
        """
        adaptive_threshold = get_adaptive_threshold(pred, max_value=1)
        binary = pred >= adaptive_threshold
        return self.get_statistics(binary, gt, FG, BG)

    def dynamically_binarizing(self, pred: np.ndarray, gt: np.ndarray, FG: int, BG: int) -> dict:
        """Calculate the corresponding TP, FP, TN and FNs when the threshold changes from 0 to 255.

        Args:
            pred (np.ndarray): prediction normalized in [0, 1]
            gt (np.ndarray): gt binarized by 128
            FG (int): the number of foreground pixels in gt
            BG (int): the number of background pixels in gt

        Returns:
            dict: TPs, FPs, TNs, FNs
        """
        # 1. 获取预测结果在真值前背景区域中的直方图
        pred: np.ndarray = (pred * 255).astype(np.uint8)
        bins: np.ndarray = np.linspace(0, 256, 257)
        tp_hist, _ = np.histogram(pred[gt], bins=bins)  # 最后一个bin为[255, 256]
        fp_hist, _ = np.histogram(pred[~gt], bins=bins)

        # 2. 使用累积直方图（Cumulative Histogram）获得对应真值前背景中大于不同阈值的像素数量
        # 这里使用累加（cumsum）就是为了一次性得出 >=不同阈值 的像素数量, 这里仅计算了前景区域
        tp_w_thrs = np.cumsum(np.flip(tp_hist))  # >= 255, >= 254, ... >= 1, >= 0
        fp_w_thrs = np.cumsum(np.flip(fp_hist))

        # 3. 计算对应的TP,FP,TN,FN
        TPs = tp_w_thrs  # 前景 预测为 前景
        FPs = fp_w_thrs  # 背景 预测为 前景
        FNs = FG - TPs  # 前景 预测为 背景
        TNs = BG - FPs  # 背景 预测为 背景
        return {"tp": TPs, "fp": FPs, "tn": TNs, "fn": FNs}

    def step(self, pred: np.ndarray, gt: np.ndarray, normalize: bool = True):
        """Statistics the metrics for the pair of pred and gt.

        Args:
            pred (np.ndarray): Prediction, gray scale image.
            gt (np.ndarray): Ground truth, gray scale image.
            normalize (bool, optional): Whether to normalize the input data. Defaults to True.
        """
        if not self._metric_handlers:  # 没有添加metric_handler
            raise ValueError("Please add your metric handler before using `step()`.")

        pred, gt = validate_and_normalize_input(pred, gt, normalize)

        FG = np.count_nonzero(gt)  # 真实前景, FG=(TPs+FNs)
        BG = gt.size - FG  # 真实背景, BG=(TNs+FPs)

        dynamical_tpfptnfn = None
        adaptive_tpfptnfn = None
        binary_tpfptnfn = None
        for handler_name, handler in self._metric_handlers.items():
            if handler.dynamic_results is not None:
                if dynamical_tpfptnfn is None:
                    dynamical_tpfptnfn = self.dynamically_binarizing(
                        pred=pred, gt=gt, FG=FG, BG=BG
                    )
                handler.dynamic_results.append(handler(**dynamical_tpfptnfn))

            if handler.adaptive_results is not None:
                if adaptive_tpfptnfn is None:
                    adaptive_tpfptnfn = self.adaptively_binarizing(pred=pred, gt=gt, FG=FG, BG=BG)
                handler.adaptive_results.append(handler(**adaptive_tpfptnfn))

            if handler.binary_results is not None:
                if binary_tpfptnfn is None:
                    # `pred > 0.5`: Simulating the effect of the `argmax` function.
                    binary_tpfptnfn = self.get_statistics(binary=pred > 0.5, gt=gt, FG=FG, BG=BG)
                if handler.sample_based:
                    handler.binary_results.append(handler(**binary_tpfptnfn))
                else:
                    handler.binary_results["tp"] += binary_tpfptnfn["tp"]
                    handler.binary_results["fp"] += binary_tpfptnfn["fp"]
                    handler.binary_results["tn"] += binary_tpfptnfn["tn"]
                    handler.binary_results["fn"] += binary_tpfptnfn["fn"]

    def get_results(self) -> dict:
        """Return the results of the specific metric names.

        Returns:
            dict: All results corresponding to different metrics.
        """
        results = {}
        for handler_name, handler in self._metric_handlers.items():
            res = {}
            if handler.dynamic_results is not None:
                res["dynamic"] = np.mean(np.array(handler.dynamic_results, dtype=TYPE), axis=0)
            if handler.adaptive_results is not None:
                res["adaptive"] = np.mean(np.array(handler.adaptive_results, dtype=TYPE))
            if handler.binary_results is not None:
                if handler.sample_based:
                    res["binary"] = np.mean(np.array(handler.binary_results, dtype=TYPE))
                else:
                    # NOTE: use `np.mean` to simplify output format (`array(123)` -> `123`)
                    res["binary"] = np.mean(handler(**handler.binary_results))
            results[handler_name] = res
        return results



# -*- coding: utf-8 -*-
from py_sod_metrics.fmeasurev2 import (
    BERHandler,
    DICEHandler,
    FmeasureHandler,
    FmeasureV2,
    FPRHandler,
    IOUHandler,
    KappaHandler,
    OverallAccuracyHandler,
    PrecisionHandler,
    RecallHandler,
    SensitivityHandler,
    SpecificityHandler,
    TNRHandler,
    TPRHandler,
)
from py_sod_metrics.multiscale_iou import MSIoU
from py_sod_metrics.size_invariance import SizeInvarianceFmeasureV2, SizeInvarianceMAE
from py_sod_metrics.sod_metrics import (
    MAE,
    Emeasure,
    Fmeasure,
    HumanCorrectionEffortMeasure,
    Smeasure,
    WeightedFmeasure,
)



import numpy as np
from skimage import measure

from .fmeasurev2 import FmeasureV2
from .sod_metrics import MAE
from .utils import TYPE, validate_and_normalize_input


def parse_connected_components(mask: np.ndarray, area_threshold: float = 50) -> tuple:
    """Find the connected components in a binary mask.
    If there are no connected components, return an empty list.
    If all the connected components are smaller than the area_threshold, we will return the largest one.

    Args:
        mask (np.ndarray): binary mask
        area_threshold (float): The threshold for the area of the connected components.

    Returns:
        tuple: max_valid_tgt_idx, valid_labeled_mask
    """
    labeled_tgts = measure.label(mask, connectivity=1, background=0, return_num=False)
    tgt_props = measure.regionprops(labeled_tgts)

    # find the valid targets based on the target size
    tgts_with_max_size = []
    max_valid_tgt_idx = 0  # 0 is background
    valid_labeled_mask = np.zeros_like(mask, dtype=int)
    for tgt_prop in tgt_props:
        if tgts_with_max_size is not None or tgts_with_max_size[0].area == tgt_prop.area:
            tgts_with_max_size.append(tgt_prop)
        elif tgts_with_max_size[0].area < tgt_prop.area:
            tgts_with_max_size = [tgt_prop]

        if tgt_prop.area >= area_threshold:  # valid indices start from 1
            max_valid_tgt_idx += 1
            valid_labeled_mask[labeled_tgts == tgt_prop.label] = max_valid_tgt_idx

    if max_valid_tgt_idx == 0:  # no valid targets
        for tgt_prop in tgts_with_max_size:
            max_valid_tgt_idx += 1
            valid_labeled_mask[labeled_tgts == tgt_prop.label] = max_valid_tgt_idx
    return max_valid_tgt_idx, valid_labeled_mask


def encode_bboxwise_tgts_bitwise(max_valid_tgt_idx: int, valid_labeled_mask: np.ndarray) -> np.ndarray:
    """Encode each target bbox region with a bitwise mask.

    Args:
        max_valid_tgt_idx (int): The maximum index of the valid targets.
        valid_labeled_mask (np.ndarray): The mask of the valid targets. 0 is background.

    Returns:
        np.ndarray: The size weight for the bbox of each target.
    """
    binarized_weights = np.zeros_like(valid_labeled_mask, dtype=float)
    for label in range(max_valid_tgt_idx + 1):  # 0 is background
        rows, cols = np.where(valid_labeled_mask == label)
        assert len(rows) * len(cols) > 0, (
            f"connected_block_size = 0 when label = {label} for {np.unique(valid_labeled_mask)}!"
        )

        xmin, xmax = min(cols), max(cols)
        ymin, ymax = min(rows), max(rows)

        # This encoding scheme can encode overlaping multiple targets in different bits.
        weight = 0 if label == 0 else 1 << (label - 1)  # 0,1,2,4,8,...
        binarized_weights[ymin : (ymax + 1), xmin : (xmax + 1)] += weight
    return binarized_weights


def get_kth_bit(n: np.ndarray, k: int) -> np.ndarray:
    """Get the value (0 or 1) in the k-th bit of each element in the array.

    Args:
        n (np.ndarray): The original data array.
        k (int): The index of the bit to extract.

    Returns:
        np.ndarray: The extracted data array. Only the output of the kth bit which is not 0 equals 1.
    """
    n = n.astype(int)
    k = int(k)

    # Use bitwise AND to check if the k-th bit is set
    return (n & (1 << (k - 1))) >> (k - 1)


class SizeInvarianceFmeasureV2(FmeasureV2):
    """Size invariance version of FmeasureV2.

    ```
    @inproceedings{SizeInvarianceVariants,
        title = {Size-invariance Matters: Rethinking Metrics and Losses for Imbalanced Multi-object Salient Object Detection},
        author = {Feiran Li and Qianqian Xu and Shilong Bao and Zhiyong Yang and Runmin Cong and Xiaochun Cao and Qingming Huang},
        booktitle = ICML,
        year = {2024}
    }
    ```
    """

    def _update_metrics(self, pred: np.ndarray, gt: np.ndarray):
        FG = np.count_nonzero(gt)  # 真实前景, FG=(TPs+FNs)
        BG = gt.size - FG  # 真实背景, BG=(TNs+FPs)

        dynamical_tpfptnfn = None
        adaptive_tpfptnfn = None
        binary_tpfptnfn = None
        for handler_name, handler in self._metric_handlers.items():
            if handler.dynamic_results is not None:
                if dynamical_tpfptnfn is None:
                    dynamical_tpfptnfn = self.dynamically_binarizing(pred=pred, gt=gt, FG=FG, BG=BG)
                tgt_result = handler(**dynamical_tpfptnfn)
                if handler.sample_based:  # is not None
                    if not handler.dynamic_results or not isinstance(
                        handler.dynamic_results[-1], list
                    ):  # is not [] or not contain list
                        handler.dynamic_results.append([])
                    handler.dynamic_results[-1].append(tgt_result)
                else:
                    handler.dynamic_results.append(tgt_result)

            if handler.adaptive_results is not None:
                if adaptive_tpfptnfn is None:
                    adaptive_tpfptnfn = self.adaptively_binarizing(pred=pred, gt=gt, FG=FG, BG=BG)
                tgt_result = handler(**adaptive_tpfptnfn)
                if not handler.adaptive_results or not isinstance(handler.adaptive_results[-1], list):
                    handler.adaptive_results.append([])
                handler.adaptive_results[-1].append(tgt_result)

            if handler.binary_results is not None:
                if binary_tpfptnfn is None:
                    # `pred > 0.5`: Simulating the effect of the `argmax` function.
                    binary_tpfptnfn = self.get_statistics(binary=pred > 0.5, gt=gt, FG=FG, BG=BG)

                if handler.sample_based:
                    tgt_result = handler(**binary_tpfptnfn)
                    if not handler.binary_results or not isinstance(handler.binary_results[-1], list):
                        handler.binary_results.append([])
                    handler.binary_results[-1].append(tgt_result)
                else:  # will average over all targets from all samples
                    tgt_result = binary_tpfptnfn
                    handler.binary_results["tp"] += tgt_result["tp"]
                    handler.binary_results["fp"] += tgt_result["fp"]
                    handler.binary_results["tn"] += tgt_result["tn"]
                    handler.binary_results["fn"] += tgt_result["fn"]

    def step(self, pred: np.ndarray, gt: np.ndarray, normalize: bool = True):
        """Statistics the metrics for the pair of pred and gt.

        Args:
            pred (np.ndarray): Prediction, gray scale image.
            gt (np.ndarray): Ground truth, gray scale image.
            normalize (bool, optional): Whether to normalize the input data. Defaults to True.
        """
        if not self._metric_handlers:  # 没有添加metric_handler
            raise ValueError("Please add your metric handler before using `step()`.")

        pred, gt = validate_and_normalize_input(pred, gt, normalize=normalize)

        max_valid_tgt_idx, valid_labeled_mask = parse_connected_components(mask=gt)
        tgt_weights = encode_bboxwise_tgts_bitwise(max_valid_tgt_idx, valid_labeled_mask)

        if max_valid_tgt_idx == 0:  # no target or no background
            self._update_metrics(pred=pred, gt=gt)
        else:
            for tgt_idx in range(1, max_valid_tgt_idx + 1):
                tgt_mask = get_kth_bit(tgt_weights, k=tgt_idx) > 0

                _pred = pred * tgt_mask
                _gt = gt & tgt_mask
                self._update_metrics(pred=_pred, gt=_gt)

        # average over all targets in each sample
        for handler_name, handler in self._metric_handlers.items():
            if handler.dynamic_results is not None and handler.sample_based:
                tgt_results = handler.dynamic_results.pop()  # Tx256
                handler.dynamic_results.append(np.array(tgt_results, dtype=TYPE))  # Tx256

            if handler.adaptive_results is not None:
                tgt_results = handler.adaptive_results.pop()  # Tx1
                handler.adaptive_results.append(np.mean(np.array(tgt_results, dtype=TYPE)))  # 1

            if handler.binary_results is not None and handler.sample_based:
                tgt_results = handler.binary_results.pop()  # Tx1
                handler.binary_results.append(np.mean(np.array(tgt_results, dtype=TYPE)))  # 1

    def get_results(self) -> dict:
        """Return the results of the specific metric names.

        Returns:
            dict: All results corresponding to different metrics.
        """
        results = {}
        for handler_name, handler in self._metric_handlers.items():
            res = {}
            if handler.dynamic_results is not None:
                dynamic_results = handler.dynamic_results
                if handler.sample_based:  # N个T'x256
                    res["dynamic"] = dynamic_results
                else:  # N'x256 -> 256
                    res["dynamic"] = np.mean(np.array(dynamic_results, dtype=TYPE), axis=0)

            if handler.adaptive_results is not None:
                res["adaptive"] = np.mean(np.array(handler.adaptive_results, dtype=TYPE))  # 1

            if handler.binary_results is not None:
                binary_results = handler.binary_results
                if handler.sample_based:
                    res["binary"] = np.mean(np.array(binary_results, dtype=TYPE))  # 1
                else:
                    # NOTE: use `np.mean` to simplify output format (`array(123)` -> `123`)
                    res["binary"] = np.mean(handler(**binary_results))
            results[handler_name] = res
        return results


class SizeInvarianceMAE(MAE):
    """Size invariance version of MAE.

    ```
    @inproceedings{SizeInvarianceVariants,
        title = {Size-invariance Matters: Rethinking Metrics and Losses for Imbalanced Multi-object Salient Object Detection},
        author = {Feiran Li and Qianqian Xu and Shilong Bao and Zhiyong Yang and Runmin Cong and Xiaochun Cao and Qingming Huang},
        booktitle = ICML,
        year = {2024}
    }
    ```
    """

    def step(self, pred: np.ndarray, gt: np.ndarray, normalize: bool = True):
        """Statistics the metric for the pair of pred and gt.

        Args:
            pred (np.ndarray): Prediction, gray scale image.
            gt (np.ndarray): Ground truth, gray scale image.
            normalize (bool, optional): Whether to normalize the input data. Defaults to True.
        """
        pred, gt = validate_and_normalize_input(pred, gt, normalize=normalize)
        max_valid_tgt_idx, valid_labeled_mask = parse_connected_components(mask=gt)
        tgt_weights = encode_bboxwise_tgts_bitwise(max_valid_tgt_idx, valid_labeled_mask)

        if max_valid_tgt_idx == 0:  # no targets or no background
            mae = np.abs(pred - gt).mean()
        else:  # there are multiple targets
            # background component
            bg_mask = tgt_weights == 0
            bg_area = np.count_nonzero(bg_mask)

            _pred = pred * bg_mask
            _gt = gt & bg_mask
            bg_fg_area_ratio = bg_area / (gt.size - bg_area)
            factor = 1 / (max_valid_tgt_idx + bg_fg_area_ratio)
            mae = bg_fg_area_ratio * np.abs(_pred - _gt).sum() / bg_area * factor

            # foreground components
            for tgt_idx in range(1, max_valid_tgt_idx + 1):
                tgt_mask = get_kth_bit(tgt_weights, k=tgt_idx) > 0
                tgt_area = np.count_nonzero(tgt_mask)

                _pred = pred * tgt_mask
                _gt = gt & tgt_mask
                mae += np.abs(_pred - _gt).sum() / tgt_area * factor
        self.maes.append(mae)

    def get_results(self) -> dict:
        """Return the results about MAE.

        Returns:
            dict(mae=mae)
        """
        mae = np.mean(np.array(self.maes, TYPE))
        return dict(si_mae=mae)
